﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml.Linq;
using SqlDeployConsole;

namespace SqlDeployTool
{
    public partial class AdvancedForm : Form
    {
        protected XElement xElement = null;
        public AdvancedForm(XElement xElement = null)
        {
            InitializeComponent();
            this.xElement = xElement;
        }

        private void AdvancedForm_Load(object sender, EventArgs e)
        {
            RenderTree(this.tree_Scripts, this.xElement);
        }

        /// <summary>
        /// Renders the tree.
        /// </summary>
        /// <param name="tree">The tree.</param>
        /// <param name="xElement">The x element.</param>
        private void RenderTree(TreeView tree, XElement xElement)
        {
            tree.BeginUpdate();
            tree.Nodes.Clear();

            if (xElement != null)
            {
                foreach (var one in xElement.Elements())
                {
                    var addedNode = tree.Nodes.Add(one.Name.LocalName);
                    List<string> scriptPaths = new List<string>();
                    foreach (var item in one.Elements(Constants.Item))
                    {
                        if (!string.IsNullOrWhiteSpace(item.Value))
                        {
                            scriptPaths.Add(item.Value);
                        }
                    }
                    RenderTree(addedNode, scriptPaths);
                }
            }
            tree.EndUpdate();
        }

        private void RenderTree(TreeNode treeNode, List<string> scriptPaths)
        {
            if (treeNode != null && scriptPaths != null)
            {
                Dictionary<string, List<string>> subPaths = new Dictionary<string, List<string>>();
                List<string> files = new List<string>();
                foreach (var one in scriptPaths)
                {
                    var pathString = one.Trim('\\');
                    int index = pathString.IndexOf('\\');
                    if (index > -1)
                    {
                        string nodeName = pathString.Substring(0, index);
                        if (!subPaths.ContainsKey(nodeName))
                        {
                            subPaths.Add(nodeName, new List<string>());
                        }
                        subPaths[nodeName].Add(pathString.Substring(index));
                    }
                    else
                    {
                        files.Add(pathString);
                    }
                }

                foreach (string key in subPaths.Keys)
                {
                    var addedNode = treeNode.Nodes.Add(key);
                    RenderTree(addedNode, subPaths[key]);
                }

                foreach (var file in files)
                {
                    treeNode.Nodes.Add(file);
                }
            }
        }

        /// <summary>
        /// Handles the Click event of the btn_Save control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btn_Save_Click(object sender, EventArgs e)
        {
            MainForm.IncludedScriptXml = GenerateXml(this.tree_Scripts);
            this.Close();
        }

        /// <summary>
        /// Handles the Click event of the btn_Cancel control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btn_Cancel_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Generates the XML.
        /// </summary>
        /// <param name="treeView">The tree view.</param>
        /// <returns></returns>
        protected XElement GenerateXml(TreeView treeView)
        {
            XElement result = null;

            if (treeView != null && treeView.Nodes != null)
            {
                result = XElement.Parse(string.Format(Constants.XmlNodeFormat, this.xElement.Name.LocalName));

                foreach (TreeNode sectionNode in treeView.Nodes)
                {
                    if (sectionNode.Checked)
                    {
                        XElement section = XElement.Parse(string.Format(Constants.XmlNodeFormat, sectionNode.Name));
                        if (sectionNode.Nodes != null)
                        {
                            foreach (TreeNode one in sectionNode.Nodes)
                            {
                                var list = GetPathsByNode(string.Empty, one);
                                foreach (var path in list)
                                {
                                    XElement element = XElement.Parse(string.Format(Constants.XmlNodeFormat, Constants.Item));
                                    element.Value = path;
                                    section.Add(element);
                                }
                            }
                        }
                        result.Add(section);
                    }
                }
            }

            return result;
        }

        /// <summary>
        /// Gets the paths by node.
        /// </summary>
        /// <param name="basePath">The base path.</param>
        /// <param name="node">The node.</param>
        /// <returns></returns>
        protected List<string> GetPathsByNode(string basePath, TreeNode node)
        {
            List<string> result = new List<string>();
            if (node != null && node.Checked)
            {
                if (string.IsNullOrWhiteSpace(basePath))
                {
                    basePath = string.Empty;
                }

                basePath = Path.Combine(basePath, node.Name);
                if (node.Nodes != null && node.Nodes.Count > 0)
                {
                    foreach (TreeNode one in node.Nodes)
                    {
                        result.AddRange(GetPathsByNode(basePath, one));
                    }
                }
                else
                {
                    result.Add(basePath);
                }
            }

            return result;
        }

        /// <summary>
        /// Handles the AfterCheck event of the tree_Scripts control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="TreeViewEventArgs"/> instance containing the event data.</param>
        private void tree_Scripts_AfterCheck(object sender, TreeViewEventArgs e)
        {
            if (e.Action != TreeViewAction.Unknown)
            {
                bool isChecked = e.Node.Checked;
                TreeView treeInstance = sender as TreeView;
                if (treeInstance != null)
                {
                    treeInstance.BeginUpdate();

                    CheckAllSubItems(e.Node, isChecked);
                    CheckAllParentItems(e.Node, isChecked);

                    treeInstance.EndUpdate();
                }
            }
        }

        /// <summary>
        /// Checks all sub items.
        /// </summary>
        /// <param name="node">The node.</param>
        /// <param name="isChecked">if set to <c>true</c> [is checked].</param>
        private void CheckAllSubItems(TreeNode node, bool isChecked)
        {
            if (node != null)
            {
                node.Checked = isChecked;
                if (node.Nodes != null && node.Nodes.Count > 0)
                {
                    foreach (TreeNode one in node.Nodes)
                    {
                        CheckAllSubItems(one, isChecked);
                    }
                }
            }
        }

        /// <summary>
        /// Checks all parent items.
        /// </summary>
        /// <param name="node">The node.</param>
        /// <param name="isChecked">if set to <c>true</c> [is checked].</param>
        private void CheckAllParentItems(TreeNode node, bool isChecked)
        {
            if (node != null)
            {
                TreeNode parentNode = node.Parent;
                while (parentNode != null)
                {
                    if (isChecked)
                    {
                        if (IsAllSubItemsChecked(parentNode))
                        {
                            parentNode.Checked = true;
                        }
                    }
                    else
                    {
                        parentNode.Checked = false;
                    }

                    parentNode = parentNode.Parent;
                }
            }
        }

        /// <summary>
        /// Determines whether [is all sub items checked] [the specified node].
        /// </summary>
        /// <param name="node">The node.</param>
        /// <returns>
        ///   <c>true</c> if [is all sub items checked] [the specified node]; otherwise, <c>false</c>.
        /// </returns>
        private bool IsAllSubItemsChecked(TreeNode node)
        {
            bool result = true; ;

            if (node != null)
            {
                if (node.Nodes != null && node.Nodes.Count > 0)
                {
                    foreach (TreeNode subNode in node.Nodes)
                    {
                        result = result & IsAllSubItemsChecked(subNode);
                        if (!result)
                        {
                            break;
                        }
                    }
                }
                else
                {
                    result = node.Checked;
                }
            }

            return result;
        }
    }
}
