﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml.Linq;
using SqlDeployConsole;
using SqlDeployConsole.Core;

namespace SqlDeployTool
{
    /// <summary>
    /// 
    /// </summary>
    public partial class AdvancedForm : Form
    {
        protected XElement configurationXml = null;
        protected XElement includedXml = null;
        protected Version baseVersion = null;
        protected Version targetVersion = null;
        protected object treeLocker = new object();

        /// <summary>
        /// Initializes a new instance of the <see cref="AdvancedForm"/> class.
        /// </summary>
        /// <param name="includedXml">The included XML.</param>
        /// <param name="configurationXml">The configuration XML.</param>
        public AdvancedForm(XElement includedXml, XElement configurationXml = null)
        {
            InitializeComponent();
            this.configurationXml = configurationXml;
            this.includedXml = includedXml;

            if (configurationXml != null)
            {
                XAttribute attribute = configurationXml.Attribute(Constants.BaseVersion);
                if (attribute != null && !string.IsNullOrWhiteSpace(attribute.Value))
                {
                    baseVersion = new Version(attribute.Value);
                }

                attribute = configurationXml.Attribute(Constants.TargetVersion);
                if (attribute != null & !string.IsNullOrWhiteSpace(attribute.Value))
                {
                    targetVersion = new Version(attribute.Value);
                }
            }
        }

        /// <summary>
        /// Handles the Load event of the AdvancedForm control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void AdvancedForm_Load(object sender, EventArgs e)
        {
            this.label_Loading.Visible = false;
            RenderTree(this.tree_Scripts, this.includedXml, this.configurationXml);
        }

        /// <summary>
        /// Renders the tree.
        /// </summary>
        /// <param name="tree">The tree.</param>
        /// <param name="includedXml">The included XML.</param>
        /// <param name="configurationXml">The x element.</param>
        private void RenderTree(TreeView tree, XElement includedXml, XElement configurationXml)
        {
            lock (treeLocker)
            {
                SwitchMask(true);

                tree.BeginUpdate();
                tree.Nodes.Clear();

                if (includedXml != null && configurationXml != null)
                {
                    foreach (var one in configurationXml.Elements())
                    {
                        var addedNode = tree.Nodes.Add(one.Name.LocalName);
                        List<string> scriptPaths = new List<string>();
                        foreach (var item in one.Elements(Constants.Item))
                        {
                            if (!string.IsNullOrWhiteSpace(item.Value))
                            {
                                scriptPaths.Add(item.Value);
                            }
                        }
                        RenderTree(includedXml.Element(one.Name.LocalName), addedNode, scriptPaths);
                    }
                }
                tree.EndUpdate();

                SwitchMask(false);
            }
        }

        /// <summary>
        /// Renders the tree.
        /// </summary>
        /// <param name="includedXml">The included XML.</param>
        /// <param name="treeNode">The tree node.</param>
        /// <param name="scriptPaths">The script paths.</param>
        private void RenderTree(XElement includedXml, TreeNode treeNode, List<string> scriptPaths)
        {
            if (treeNode != null && scriptPaths != null)
            {
                Dictionary<string, List<string>> subPaths = new Dictionary<string, List<string>>();
                List<string> files = new List<string>();
                foreach (var one in scriptPaths)
                {
                    var pathString = one.Trim('\\');
                    int index = pathString.IndexOf('\\');
                    if (index > -1)
                    {
                        string nodeName = pathString.Substring(0, index);
                        if (!subPaths.ContainsKey(nodeName))
                        {
                            subPaths.Add(nodeName, new List<string>());
                        }
                        subPaths[nodeName].Add(pathString.Substring(index));
                    }
                    else
                    {
                        files.Add(pathString);
                    }
                }

                foreach (string key in subPaths.Keys)
                {
                    var addedNode = treeNode.Nodes.Add(key);
                    RenderTree(includedXml, addedNode, subPaths[key]);
                }

                foreach (var file in files)
                {
                    var addedNode = treeNode.Nodes.Add(file);
                    addedNode.Checked = ScriptHelper.HasMatchedValueNode(GetCombinedNodeValue(addedNode), includedXml);
                    CheckAllParentItems(addedNode, true);
                }
            }
        }

        /// <summary>
        /// Handles the Click event of the btn_Save control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btn_Save_Click(object sender, EventArgs e)
        {
            MainForm.IncludedScriptXml = GenerateXml(this.tree_Scripts);
            this.Close();
        }

        /// <summary>
        /// Handles the Click event of the btn_Cancel control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btn_Cancel_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Generates the XML.
        /// </summary>
        /// <param name="treeView">The tree view.</param>
        /// <returns></returns>
        protected XElement GenerateXml(TreeView treeView)
        {
            XElement result = null;

            if (treeView != null && treeView.Nodes != null)
            {
                result = XElement.Parse(string.Format(Constants.XmlNodeFormat, this.configurationXml.Name.LocalName));
                result.SetAttributeValue(Constants.BaseVersion, this.baseVersion == null ? string.Empty : this.baseVersion.ToString());
                result.SetAttributeValue(Constants.TargetVersion, this.targetVersion == null ? string.Empty : this.targetVersion.ToString());
                foreach (TreeNode sectionNode in treeView.Nodes)
                {
                    XElement section = XElement.Parse(string.Format(Constants.XmlNodeFormat, sectionNode.Text));
                    if (sectionNode.Nodes != null)
                    {
                        foreach (TreeNode one in sectionNode.Nodes)
                        {
                            var list = GetPathsByNode(string.Empty, one);
                            foreach (var path in list)
                            {
                                XElement element = XElement.Parse(string.Format(Constants.XmlNodeFormat, Constants.Item));
                                element.Value = path;
                                section.Add(element);
                            }
                        }
                    }
                    result.Add(section);
                }
            }

            return result;
        }

        /// <summary>
        /// Gets the paths by node.
        /// </summary>
        /// <param name="basePath">The base path.</param>
        /// <param name="node">The node.</param>
        /// <returns></returns>
        protected List<string> GetPathsByNode(string basePath, TreeNode node)
        {
            List<string> result = new List<string>();
            if (node != null && (node.Checked || (node.Nodes != null && node.Nodes.Count > 0)))
            {
                if (string.IsNullOrWhiteSpace(basePath))
                {
                    basePath = string.Empty;
                }

                basePath = Path.Combine(basePath, node.Text);
                if (node.Nodes != null && node.Nodes.Count > 0)
                {
                    foreach (TreeNode one in node.Nodes)
                    {
                        result.AddRange(GetPathsByNode(basePath, one));
                    }
                }
                else
                {
                    result.Add(basePath);
                }
            }

            return result;
        }

        /// <summary>
        /// Handles the AfterCheck event of the tree_Scripts control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="TreeViewEventArgs"/> instance containing the event data.</param>
        private void tree_Scripts_AfterCheck(object sender, TreeViewEventArgs e)
        {
            if (e.Action != TreeViewAction.Unknown)
            {
                lock (treeLocker)
                {
                    SwitchMask(true);

                    bool isChecked = e.Node.Checked;
                    TreeView treeInstance = sender as TreeView;
                    if (treeInstance != null)
                    {
                        treeInstance.BeginUpdate();

                        CheckAllSubItems(e.Node, isChecked);
                        CheckAllParentItems(e.Node, isChecked);

                        treeInstance.EndUpdate();
                    }

                    SwitchMask(false);
                }
            }
        }

        /// <summary>
        /// Checks all sub items.
        /// </summary>
        /// <param name="node">The node.</param>
        /// <param name="isChecked">if set to <c>true</c> [is checked].</param>
        private void CheckAllSubItems(TreeNode node, bool isChecked)
        {
            if (node != null)
            {
                node.Checked = isChecked;
                if (node.Nodes != null && node.Nodes.Count > 0)
                {
                    foreach (TreeNode one in node.Nodes)
                    {
                        CheckAllSubItems(one, isChecked);
                    }
                }
            }
        }

        /// <summary>
        /// Checks all parent items.
        /// </summary>
        /// <param name="node">The node.</param>
        /// <param name="isChecked">if set to <c>true</c> [is checked].</param>
        private void CheckAllParentItems(TreeNode node, bool isChecked)
        {
            if (node != null)
            {
                TreeNode parentNode = node.Parent;
                while (parentNode != null)
                {
                    if (isChecked)
                    {
                        if (IsAllSubItemsChecked(parentNode))
                        {
                            parentNode.Checked = true;
                        }
                    }
                    else
                    {
                        parentNode.Checked = false;
                    }

                    parentNode = parentNode.Parent;
                }
            }
        }

        /// <summary>
        /// Determines whether [is all sub items checked] [the specified node].
        /// </summary>
        /// <param name="node">The node.</param>
        /// <returns>
        ///   <c>true</c> if [is all sub items checked] [the specified node]; otherwise, <c>false</c>.
        /// </returns>
        private bool IsAllSubItemsChecked(TreeNode node)
        {
            bool result = true; ;

            if (node != null)
            {
                if (node.Nodes != null && node.Nodes.Count > 0)
                {
                    foreach (TreeNode subNode in node.Nodes)
                    {
                        result = result & IsAllSubItemsChecked(subNode);
                        if (!result)
                        {
                            break;
                        }
                    }
                }
                else
                {
                    result = node.Checked;
                }
            }

            return result;
        }

        /// <summary>
        /// Gets the combined node value.
        /// </summary>
        /// <param name="node">The node.</param>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        private string GetCombinedNodeValue(TreeNode node, string value = null)
        {
            string result = string.Empty;

            bool isFirst = value == null;
            if (isFirst)
            {
                value = string.Empty;
            }

            if (node != null && node.Parent != null)
            {
                result = GetCombinedNodeValue(node.Parent, "\\" + node.Text + value);
            }
            else
            {
                result = value;
            }

            return isFirst ? result.Trim('\\') : result;
        }

        /// <summary>
        /// Switches the mask.
        /// </summary>
        /// <param name="isVisible">if set to <c>true</c> [is visible].</param>
        private void SwitchMask(bool isVisible)
        {
            this.label_Loading.Visible = isVisible;
            this.tree_Scripts.Enabled = !this.label_Loading.Visible;
        }
    }
}
